/*
   This file is part of NavisionAutomation.

   NavisionAutomation is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   NavisionAutomation is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with NavisionAutomation.  If not, see <http://www.gnu.org/licenses/>.  */
#include <comdef.h>
#include <windows.h>
#include "NavisionAutomation.h"
#include "Navision.h"
#include <olectl.h>
#include <sstream>
#include <vector>
#include <string>
#include <fstream>
_COM_SMARTPTR_TYPEDEF(INSObjectDesigner, __uuidof(INSObjectDesigner));
_COM_SMARTPTR_TYPEDEF(INSApplication, __uuidof(INSApplication));
_COM_SMARTPTR_TYPEDEF(INSMenuButton, __uuidof(INSObjectDesigner));
_COM_SMARTPTR_TYPEDEF(INSHyperlink, __uuidof(INSHyperlink));
_COM_SMARTPTR_TYPEDEF(INSAppBase, __uuidof(INSAppBase));
_COM_SMARTPTR_TYPEDEF(INSTable, __uuidof(INSTable));
_COM_SMARTPTR_TYPEDEF(INSForm, __uuidof(INSForm));
_COM_SMARTPTR_TYPEDEF(INSRec, __uuidof(INSRec));
static HINSTANCE g_hinstDLL;
const char *g_RegTable[][3] = {
	// format is { key, value name, value }
	{ "CLSID\\{B1099FD9-F00F-4f42-8C16-C56363D801B7}", 0, "NavisionAutomation" },
	{ "CLSID\\{B1099FD9-F00F-4f42-8C16-C56363D801B7}\\InprocServer32", 0, (const char*)-1 },
	{ "CLSID\\{B1099FD9-F00F-4f42-8C16-C56363D801B7}\\InprocServer32", "ThreadingModel", "Free"},
	{ "CLSID\\{B1099FD9-F00F-4f42-8C16-C56363D801B7}\\ProgID", 0, "NavisionAutomation.1"}, 
	{ "NavisionAutomation.1", 0, "NavisionAutomation" },
	{ "NavisionAutomation.1\\CLSID", 0, "{B1099FD9-F00F-4f42-8C16-C56363D801B7}" },
};
STDAPI DllUnregisterServer(void) {
	OutputDebugString("DllUnregisterServer");
	HRESULT hr = S_OK;
	int nEntries = sizeof(g_RegTable) / sizeof(*g_RegTable);
	for (int i = nEntries - 1; i >= 0; i--) {
		const char *pszKeyName = g_RegTable[i][0];
		long err = RegDeleteKeyA(HKEY_CLASSES_ROOT, pszKeyName);
		if (err != ERROR_SUCCESS) hr = S_FALSE;
	}
	return hr;
}
STDAPI DllRegisterServer() {
	OutputDebugString("DllRegisterServer");
	HRESULT hr = S_OK;
	// look up server's file name
	char szFileName[MAX_PATH];
	GetModuleFileNameA(g_hinstDLL, szFileName, MAX_PATH);
	// register entries from table
	int nEntries = sizeof(g_RegTable)/ sizeof(*g_RegTable);
	for (int i = 0; SUCCEEDED(hr) && i < nEntries; i++) {
		const char *pszKeyName = g_RegTable[i][0];
		const char *pszValueName = g_RegTable[i][1];
		const char *pszValue = g_RegTable[i][2];
		// m,ap rogue value to module file name
		if (pszValue == (const char*)-1)
			pszValue = szFileName;
		HKEY hKey;
		// create the key
		long err = RegCreateKeyA(HKEY_CLASSES_ROOT, pszKeyName, &hKey);
		if (err == ERROR_SUCCESS) {
			//set the value
			err = RegSetValueExA(hKey, pszValueName, 0, REG_SZ, (const BYTE*)pszValue, (strlen(pszValue) + 1));
			RegCloseKey(hKey);
			OutputDebugString("RegCloseKey");
		}
		if (err != ERROR_SUCCESS) {
			// if cannot add key or value, back out and fail
			DllUnregisterServer();
			hr = SELFREG_E_CLASS;
		}
	}

	// register typelib
	wchar_t wszFileName[_MAX_PATH];
	mbstowcs(wszFileName, szFileName, MAX_PATH);
	ITypeLib *pptlib = 0;
	hr = LoadTypeLib(wszFileName, &pptlib);
	if (SUCCEEDED(hr)) {
		hr = RegisterTypeLib(pptlib, wszFileName, NULL);
	}
	return hr;
}
typedef HRESULT mydistream_run_func(void *);
typedef void mydistream_run_func_free(void *);
void mydistream_run_func_free_writeobjects(void *param) {
	OutputDebugString("mydistream_run_func_free_writeobjects");
	// FIXME
}
HRESULT mydistream_run_func_writeobjects(void *param) {
	OutputDebugString("mydistream_run_func_writeobjects");
	void **param_array = (void**)param;
	INSObjectDesigner *insobjectdesigner = static_cast<INSObjectDesigner*>(param_array[0]);
	IStream *istream = static_cast<IStream*>(param_array[1]);
	HRESULT hr = insobjectdesigner->WriteObjects(istream);
	return hr;
}
void mydistream_run_func_writeobjects_free(void *param) {
	if (param != 0) {
		void **param_array = (void**)param;
		INSObjectDesigner *insobjectdesigner = static_cast<INSObjectDesigner*>(param_array[0]);
		IStream *istream = static_cast<IStream*>(param_array[1]);
		istream->Release();
		insobjectdesigner->Release();
		free(param);
	}
}
class MyDIStream : public DIStream {
private:
	long m_cRef;
	ITypeInfo *m_pTypeInfo;
	IStreamPtr strm;
public:
	mydistream_run_func *run_func;
	mydistream_run_func_free *run_func_free;
	void *run_func_param;
	MyDIStream(IStreamPtr strm) : m_cRef(0), run_func(0), run_func_param(0), run_func_free(0) {
		ITypeLib *ptl = 0;
		HRESULT hr = LoadRegTypeLib(LIBID_NavisionAutomation, 1,0,0, &ptl);
                if (SUCCEEDED(hr)) {
			hr = ptl->GetTypeInfoOfGuid(IID_DIStream, &m_pTypeInfo);
			ptl->Release();
		}
		this->strm = strm;
	}
	~MyDIStream() {
		OutputDebugString("~MyDIStream");
		if (this->run_func_free) this->run_func_free(run_func_param);
		OutputDebugString("~MyDIStream 1");
		strm->Release();
		OutputDebugString("~MyDIStream done");
	}
	HRESULT _stdcall QueryInterface(const IID & riid,void **ppvObject) {
		if (IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_IDispatch) || IsEqualIID(riid, IID_DIStream)) {
			*ppvObject = this;
			reinterpret_cast<IUnknown*>(*ppvObject)->AddRef();
			return NOERROR;
		} else {
			*ppvObject = 0;
			return E_NOINTERFACE;
		}
	}
	ULONG _stdcall AddRef() {
		return InterlockedIncrement(&m_cRef);
	}
	ULONG _stdcall Release() {
		OutputDebugString("MyDIStream::Release");
		LONG res = InterlockedDecrement(&m_cRef);
		if (res == 0) delete this;
		return res;
	}
	HRESULT _stdcall GetTypeInfoCount(UINT * pctinfo) {
		*pctinfo = 1;
		return S_OK;
	}
	HRESULT _stdcall GetTypeInfo(UINT iTInfo, LCID lcid, ITypeInfo ** ppTInfo) {
		(*ppTInfo = m_pTypeInfo)->AddRef();
		return S_OK;
	}
	HRESULT _stdcall GetIDsOfNames(const IID & riid,LPOLESTR *rgszNames,UINT cNames,LCID lcid, DISPID* rgDispId) {
		return m_pTypeInfo->GetIDsOfNames(rgszNames, cNames, rgDispId);
	}
	HRESULT _stdcall Invoke(DISPID dispIdMember,const IID &riid,LCID lcid,WORD wFlags, DISPPARAMS * pDispParams,VARIANT *pVarResult, EXCEPINFO *pExcepInfo, UINT * puArgErr) {
		void *pvThis = static_cast<DIStream*>(this);
		return m_pTypeInfo->Invoke(pvThis, dispIdMember, wFlags, pDispParams, pVarResult, pExcepInfo, puArgErr);
	}
	HRESULT _stdcall Read(unsigned long cb, BSTR *pv) {
		ULONG cbRead;
    		char* ret = (char*)malloc(cb+1);
		LARGE_INTEGER li = {0};
		strm->Read(ret, cb, &cbRead);
		ret[cbRead] = 0;
//		---
//		std::string stringret = std::string((char*)ret, cbRead);
//		*pv = SysAllocStringLen(std::wstring(stringret.begin(), stringret.end()).c_str(), cbRead);
//		---
//		_bstr_t ret_bstr_t = _bstr_t(ret);
//		*pv = SysAllocStringLen(ret_bstr_t, cbRead);
		wchar_t * wcstring = new wchar_t[cbRead+1];
		size_t convertedChars = 0;
		mbstowcs_s(&convertedChars, wcstring, cbRead+1, ret, _TRUNCATE);
		*pv = SysAllocStringLen(wcstring, cbRead);
		free(wcstring);
		free(ret);
		return S_OK;
	}
	HRESULT _stdcall Write(BSTR pv,unsigned long cb,unsigned long *pcbWritten) {
		OutputDebugString("Write");
		_bstr_t pv_bstr_t = pv;
		strm->Write((char*)pv_bstr_t, cb, pcbWritten);
/*
		const size_t newsize = cb;
		char *mbstring = (char*)malloc(newsize);
		size_t convertedChars = 0;
		errno_t err = wcstombs_s(&convertedChars, mbstring, newsize, (wchar_t*)(pv), newsize/2);
*/
/*
		WideCharToMultiByte(CP_ACP, WC_COMPOSITECHECK, pv, newsize, mbstring, newsize, NULL, NULL);
*/
/*
		OutputDebugString("mbstring:");
		OutputDebugString(mbstring); 
		strm->Write(mbstring, cb, pcbWritten);
		free(mbstring);
*/
/*		std::ofstream f;
		f.open("c:\\temp\\write.txt");
		for (int i = 0; i < cb; i++) f << ((unsigned char*)pv)[i*2];
		f.close();
*/
/*		std::wstring pv_wstring = std::wstring(pv);
		strm->Write(std::string(pv_wstring.begin(), pv_wstring.end()).c_str(), cb, pcbWritten);
*/
		return S_OK;
	}
	HRESULT _stdcall Seek(long dlibMove, int dwOrigin,unsigned long *plibNewPosition) {
		LARGE_INTEGER mydlibMove = {dlibMove};
		ULARGE_INTEGER ret;
    		strm->Seek(mydlibMove, dwOrigin, &ret);
		*plibNewPosition = ret.LowPart;
		return S_OK;
	}
	HRESULT _stdcall SetSize(long libnewSize) {
		ULARGE_INTEGER libnewSize_ULI;
		libnewSize_ULI.LowPart = libnewSize;
		return strm->SetSize(libnewSize_ULI);
	}
	HRESULT _stdcall Commit(int grfCommitFlags) {
		return strm->Commit(grfCommitFlags);
	}
	HRESULT _stdcall Revert(void) {
		return strm->Revert();
	}
	HRESULT _stdcall LockRegion(long libOffset,long cb,int dwLockType) {
		ULARGE_INTEGER libOffset_ULI;
		ULARGE_INTEGER cb_ULI;
		libOffset_ULI.LowPart = libOffset;
		cb_ULI.LowPart = cb;
		strm->LockRegion(libOffset_ULI, cb_ULI, dwLockType);
		return S_OK;
	}
	HRESULT _stdcall UnlockRegion(long libOffset,long cb,int dwLockType) {
		ULARGE_INTEGER libOffset_ULI;
		ULARGE_INTEGER cb_ULI;
		libOffset_ULI.LowPart = libOffset;
		cb_ULI.LowPart = cb;
		strm->UnlockRegion(libOffset_ULI, cb_ULI, dwLockType);
		return S_OK;
	}
	HRESULT _stdcall Size(long *cbSize) {
		STATSTG statstg;
		strm->Stat(&statstg, 0);
		*cbSize = statstg.cbSize.LowPart;
		return S_OK;
	}
	HRESULT _stdcall Run() {
		if (run_func == NULL || run_func_param == NULL) return S_OK;
		return run_func(run_func_param);
	}
};
class MyINSCallbackEnum : public INSCallbackEnum {
private:
	long m_cRef;
public:
	INSRec* record;
	int fieldval_id, proc5_a, proc6_a, fielddef_id, fielddef_datalength, fielddef_f;
	LPOLESTR fieldval_value, fieldval_type, proc5_b, proc6_b, fielddef_name, fielddef_caption, fielddef_datatype;
	int done;
	HANDLE record_thread_event, fieldval_thread_event, proc5_thread_event, proc6_thread_event, fielddef_thread_event,
	       record_dispatch_event, fieldval_dispatch_event, proc5_dispatch_event, proc6_dispatch_event, fielddef_dispatch_event;

	MyINSCallbackEnum() : record(0), done(0), m_cRef(0) {
		OutputDebugString("MyINSCallbackEnum()");
		//threadevent = CreateEvent(NULL, FALSE, FALSE, NULL);
		//dispatchevent = CreateEvent(NULL, FALSE, FALSE, NULL);
		record_thread_event = CreateEvent(NULL, FALSE, FALSE, NULL);
		fieldval_thread_event = CreateEvent(NULL, FALSE, FALSE, NULL);
		proc5_thread_event = CreateEvent(NULL, FALSE, FALSE, NULL);
		proc6_thread_event = CreateEvent(NULL, FALSE, FALSE, NULL);
		fielddef_thread_event = CreateEvent(NULL, FALSE, FALSE, NULL);

		record_dispatch_event = CreateEvent(NULL, FALSE, FALSE, NULL);
		fieldval_dispatch_event = CreateEvent(NULL, FALSE, FALSE, NULL);
		proc5_dispatch_event = CreateEvent(NULL, FALSE, FALSE, NULL);
		proc6_dispatch_event = CreateEvent(NULL, FALSE, FALSE, NULL);
		fielddef_dispatch_event = CreateEvent(NULL, FALSE, FALSE, NULL);

	}
	~MyINSCallbackEnum() {
		OutputDebugString("~MyINSCallbackEnum");
		CloseHandle(record_thread_event);
		CloseHandle(fieldval_thread_event);
		CloseHandle(proc5_thread_event);
		CloseHandle(proc6_thread_event);
		CloseHandle(fielddef_thread_event);

		CloseHandle(record_dispatch_event);
		CloseHandle(fieldval_dispatch_event);
		CloseHandle(proc5_dispatch_event);
		CloseHandle(proc6_dispatch_event);
		CloseHandle(fielddef_dispatch_event);
	}
	HRESULT _stdcall QueryInterface(const IID & riid,void **ppvObject) {
		if (IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_INSCallbackEnum)) {
			*ppvObject = this;
			reinterpret_cast<IUnknown*>(*ppvObject)->AddRef();
			return NOERROR;
		} else {
			*ppvObject = 0;
			return E_NOINTERFACE;
		}
	}
	ULONG _stdcall AddRef() {
		return InterlockedIncrement(&m_cRef);
	}
	ULONG _stdcall Release() {
		LONG res = InterlockedDecrement(&m_cRef);
		std::stringstream ss;
		ss << "MyINSCallbackEnum::Release (" << m_cRef << ")";
		OutputDebugString(ss.str().c_str());
		if (res == 0) delete this;
		return res;
	}
	HRESULT _stdcall NextRecord(INSRec *record) {
		OutputDebugString("NextRecord");
		this->record = record;
		SetEvent(record_dispatch_event);
		WaitForSingleObject(record_thread_event, INFINITE);
		return S_OK;
	}
	HRESULT _stdcall NextFieldVal(int id, LPOLESTR value, LPOLESTR type) {
		OutputDebugString("NextFieldVal");
		this->fieldval_id = id;
		this->fieldval_value = value;
		this->fieldval_type = type;
		SetEvent(fieldval_dispatch_event);
		WaitForSingleObject(fieldval_thread_event, INFINITE);
		return S_OK;
	}
	HRESULT _stdcall proc5(int a, LPOLESTR b) {
		OutputDebugString("proc5");
		this->proc5_a = a;
		this->proc5_b = b;
		SetEvent(proc5_dispatch_event);
		WaitForSingleObject(proc5_thread_event, INFINITE);
		return S_OK;
	}
	HRESULT _stdcall proc6(int a, LPOLESTR b) {
		OutputDebugString("proc6");
		this->proc6_a = a;
		this->proc6_b = b;
		SetEvent(proc6_dispatch_event);
		WaitForSingleObject(proc6_thread_event, INFINITE);
		return S_OK;
	}
	HRESULT _stdcall NextFieldDef(int id, LPOLESTR name, LPOLESTR caption, LPOLESTR datatype, int datalength, int f) {
		OutputDebugString("thread NextFieldDef");
		this->fielddef_id = id;
		this->fielddef_name = name;
		this->fielddef_caption = caption;
		this->fielddef_datatype = datatype;
		this->fielddef_datalength = datalength;
		this->fielddef_f = f;
		SetEvent(fielddef_dispatch_event);
		WaitForSingleObject(fielddef_thread_event, INFINITE);
		return S_OK;
	}
};
class MyDIRec : public DIRec {
private:
	long m_cRef;
	ITypeInfo *m_pTypeInfo;
	INSRec* insrec;
	IMallocPtr imalloc;
public:
	MyDIRec(INSRec *insrec) : m_cRef(0) {
		ITypeLib *ptl = 0;
		HRESULT hr = LoadRegTypeLib(LIBID_NavisionAutomation, 1,0,0, &ptl);
                if (SUCCEEDED(hr)) {
			hr = ptl->GetTypeInfoOfGuid(IID_DIRec, &m_pTypeInfo);
			ptl->Release();
		}
		this->insrec = insrec;
		CoGetMalloc(1, &imalloc);
	}
	~MyDIRec() {
	}
	HRESULT _stdcall QueryInterface(const IID & riid,void **ppvObject) {
		if (IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_IDispatch) || IsEqualIID(riid, IID_DIRec)) {
			*ppvObject = this;
			reinterpret_cast<IUnknown*>(*ppvObject)->AddRef();
			return NOERROR;
		} else {
			*ppvObject = 0;
			return E_NOINTERFACE;
		}
	}
	ULONG _stdcall AddRef() {
		return InterlockedIncrement(&m_cRef);
	}
	ULONG _stdcall Release() {
		LONG res = InterlockedDecrement(&m_cRef);
		if (res == 0) delete this;
		return res;
	}
	HRESULT _stdcall GetTypeInfoCount(UINT * pctinfo) {
		*pctinfo = 1;
		return S_OK;
	}
	HRESULT _stdcall GetTypeInfo(UINT iTInfo, LCID lcid, ITypeInfo ** ppTInfo) {
		(*ppTInfo = m_pTypeInfo)->AddRef();
		return S_OK;
	}
	HRESULT _stdcall GetIDsOfNames(const IID & riid,LPOLESTR *rgszNames,UINT cNames,LCID lcid, DISPID* rgDispId) {
		return m_pTypeInfo->GetIDsOfNames(rgszNames, cNames, rgDispId);
	}
	HRESULT _stdcall Invoke(DISPID dispIdMember,const IID &riid,LCID lcid,WORD wFlags, DISPPARAMS * pDispParams,VARIANT *pVarResult, EXCEPINFO *pExcepInfo, UINT * puArgErr) {
		void *pvThis = static_cast<DIRec*>(this);
		return m_pTypeInfo->Invoke(pvThis, dispIdMember, wFlags, pDispParams, pVarResult, pExcepInfo, puArgErr);
	}
	HRESULT _stdcall proc3(int a,BSTR b,int c) {
		insrec->proc3(a, b, c);
		return S_OK;
	}
	HRESULT _stdcall GetFieldValue(int a,BSTR *b) {
		LPOLESTR ret;
		insrec->GetFieldValue(a,&ret);
		*b = SysAllocString(ret);
		imalloc->Free(ret);
		return S_OK;
	}
	HRESULT _stdcall EnumFieldValues(IDispatch **en);
};
class MyDIEnum : public DIEnum {
private:
	long m_cRef;
	ITypeInfo *m_pTypeInfo;
	int record_first, fieldval_first, proc5_first, proc6_first, fielddef_first;
public:
	MyINSCallbackEnum* myinscallbackenum;
	MyDIEnum() : m_cRef(0), record_first(1), fieldval_first(1), proc5_first(1), proc6_first(1), fielddef_first(1) {
		OutputDebugString("MyDIEnum");
		ITypeLib *ptl = 0;
		HRESULT hr = LoadRegTypeLib(LIBID_NavisionAutomation, 1,0,0, &ptl);
		myinscallbackenum = new MyINSCallbackEnum;
		myinscallbackenum->AddRef();
                if (SUCCEEDED(hr)) {
			hr = ptl->GetTypeInfoOfGuid(IID_DIEnum, &m_pTypeInfo);
			ptl->Release();
		}
	}
	~MyDIEnum() {
		OutputDebugString("~MyDIEnum");
		while (!myinscallbackenum->done) {
			OutputDebugString("SetEvents");
			SetEvent(myinscallbackenum->record_dispatch_event);
			SetEvent(myinscallbackenum->fieldval_dispatch_event);
			SetEvent(myinscallbackenum->proc5_dispatch_event);
			SetEvent(myinscallbackenum->proc6_dispatch_event);
			SetEvent(myinscallbackenum->fielddef_dispatch_event);
			Sleep(1000);
		}
		myinscallbackenum->Release();
	}
	HRESULT _stdcall QueryInterface(const IID & riid,void **ppvObject) {
		if (IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_IDispatch) || IsEqualIID(riid, IID_DIEnum)) {
			*ppvObject = this;
			reinterpret_cast<IUnknown*>(*ppvObject)->AddRef();
			return NOERROR;
		} else {
			*ppvObject = 0;
			return E_NOINTERFACE;
		}
	}
	ULONG _stdcall AddRef() {
		return InterlockedIncrement(&m_cRef);
	}
	ULONG _stdcall Release() {
		LONG res = InterlockedDecrement(&m_cRef);
		std::stringstream ss;
		ss << "MyDIEnum::Release (" << res << ")";
		OutputDebugString(ss.str().c_str());
		if (res == 0) delete this;
		return res;
	}
	HRESULT _stdcall GetTypeInfoCount(UINT * pctinfo) {
		*pctinfo = 1;
		return S_OK;
	}
	HRESULT _stdcall GetTypeInfo(UINT iTInfo, LCID lcid, ITypeInfo ** ppTInfo) {
		(*ppTInfo = m_pTypeInfo)->AddRef();
		return S_OK;
	}
	HRESULT _stdcall GetIDsOfNames(const IID & riid,LPOLESTR *rgszNames,UINT cNames,LCID lcid, DISPID* rgDispId) {
		return m_pTypeInfo->GetIDsOfNames(rgszNames, cNames, rgDispId);
	}
	HRESULT _stdcall Invoke(DISPID dispIdMember,const IID &riid,LCID lcid,WORD wFlags, DISPPARAMS * pDispParams,VARIANT *pVarResult, EXCEPINFO *pExcepInfo, UINT * puArgErr) {
		void *pvThis = static_cast<DIEnum*>(this);
		return m_pTypeInfo->Invoke(pvThis, dispIdMember, wFlags, pDispParams, pVarResult, pExcepInfo, puArgErr);
	}
	HRESULT _stdcall NextRecord(IDispatch **record) {
		if (record_first) record_first = 0;
		else SetEvent(myinscallbackenum->record_thread_event);
		if (myinscallbackenum->done) return S_OK;
		WaitForSingleObject(myinscallbackenum->record_dispatch_event, INFINITE);
		if (myinscallbackenum->done) return S_OK;
		*record = new MyDIRec(myinscallbackenum->record);
		(*record)->AddRef();
		return S_OK;
	}
	HRESULT _stdcall NextFieldVal(int *id,BSTR *value,BSTR *type) {
		if (fieldval_first) fieldval_first = 0;
		else SetEvent(myinscallbackenum->fieldval_thread_event);
		if (myinscallbackenum->done) return S_OK;
		WaitForSingleObject(myinscallbackenum->fieldval_dispatch_event, INFINITE);
		if (myinscallbackenum->done) return S_OK;
		*id = myinscallbackenum->fieldval_id;
		*value = myinscallbackenum->fieldval_value;
		*type = myinscallbackenum->fieldval_type;
		return S_OK;
	}
	HRESULT _stdcall proc5(int *a,BSTR *b) {
		if (proc5_first) proc5_first = 0;
		else SetEvent(myinscallbackenum->proc5_thread_event);
		if (myinscallbackenum->done) return S_OK;
		WaitForSingleObject(myinscallbackenum->proc5_dispatch_event, INFINITE);
		if (myinscallbackenum->done) return S_OK;
		*a = myinscallbackenum->proc5_a;
		*b = SysAllocString(myinscallbackenum->proc5_b);
		return S_OK;
	}
	HRESULT _stdcall proc6(int *a,BSTR *b) {
		if (proc6_first) proc6_first = 0;
		else SetEvent(myinscallbackenum->proc6_thread_event);
		if (myinscallbackenum->done) return S_OK;
		WaitForSingleObject(myinscallbackenum->proc6_dispatch_event, INFINITE);
		if (myinscallbackenum->done) return S_OK;
		*a = myinscallbackenum->proc6_a;
		*b = SysAllocString(myinscallbackenum->proc6_b);
		return S_OK;
	}
	HRESULT _stdcall NextFieldDef(int *id,BSTR *name,BSTR *caption,BSTR *datatype,int *datalength,int *f) {
		OutputDebugString("dispatch NextFieldDef");
		if (fielddef_first) fielddef_first = 0;
		else SetEvent(myinscallbackenum->fielddef_thread_event);
		if (myinscallbackenum->done) return S_OK;
		WaitForSingleObject(myinscallbackenum->fielddef_dispatch_event, INFINITE);
		if (myinscallbackenum->done) return S_OK;
		*id = myinscallbackenum->fielddef_id;
		*name = SysAllocString(myinscallbackenum->fielddef_name);
		*caption = SysAllocString(myinscallbackenum->fielddef_caption);
		*datatype = SysAllocString(myinscallbackenum->fielddef_datatype);
		*datalength = myinscallbackenum->fielddef_datalength;
		*f = myinscallbackenum->fielddef_f;
		return S_OK;
	}
};
DWORD WINAPI callback_insrec_EnumFieldValues(LPVOID lpParameter) {
	OutputDebugString("callback_insrec_EnumFieldValues");
	HRESULT hr = CoInitializeEx(0, COINIT_MULTITHREADED);
	if (SUCCEEDED(hr)) OutputDebugString("CoInitializeEx SUCCEEDED");
	else OutputDebugString("CoInitializeEx FAILED");
			
	INSRec* insrec = static_cast<INSRec*>(((void**)lpParameter)[0]);
	MyINSCallbackEnum *myinscallbackenum = static_cast<MyINSCallbackEnum*>(((void**)lpParameter)[1]);
	hr = insrec->EnumFieldValues(myinscallbackenum);
	insrec->Release();
	OutputDebugString("instable->EnumFieldValues done!");
	SetEvent(myinscallbackenum->record_dispatch_event);
	SetEvent(myinscallbackenum->fieldval_dispatch_event);
	SetEvent(myinscallbackenum->proc5_dispatch_event);
	SetEvent(myinscallbackenum->proc6_dispatch_event);
	SetEvent(myinscallbackenum->fielddef_dispatch_event);
	myinscallbackenum->Release();
	free(lpParameter);
	return hr;
}
DWORD WINAPI callback_instable_EnumRecords(LPVOID lpParameter) {
	OutputDebugString("callback_instable_EnumRecords");
	HRESULT hr = CoInitializeEx(0, COINIT_MULTITHREADED);
	if (SUCCEEDED(hr)) OutputDebugString("CoInitializeEx SUCCEEDED");
	else OutputDebugString("CoInitializeEx FAILED");
			
	INSTable* instable = static_cast<INSTable*>(((void**)lpParameter)[0]);
	MyINSCallbackEnum *myinscallbackenum = static_cast<MyINSCallbackEnum*>(((void**)lpParameter)[1]);
	OutputDebugString("free(lpParameter)");
	free(lpParameter);
	hr = instable->EnumRecords(myinscallbackenum);
	OutputDebugString("myinscallbackenum->done = 1");
	myinscallbackenum->done = 1;
	OutputDebugString("SetEvents");
	SetEvent(myinscallbackenum->record_dispatch_event);
	SetEvent(myinscallbackenum->fieldval_dispatch_event);
	SetEvent(myinscallbackenum->proc5_dispatch_event);
	SetEvent(myinscallbackenum->proc6_dispatch_event);
	SetEvent(myinscallbackenum->fielddef_dispatch_event);
	OutputDebugString("myinscallbackenum->Release()");
	myinscallbackenum->Release();
	OutputDebugString("callback_instable_EnumRecords done");
	return hr;
}
DWORD WINAPI callback_insappbase_proc7(LPVOID lpParameter) {
	OutputDebugString("callback_insappbase_proc7");
	HRESULT hr = CoInitializeEx(0, COINIT_MULTITHREADED);
	if (SUCCEEDED(hr)) OutputDebugString("CoInitializeEx SUCCEEDED");
	else OutputDebugString("CoInitializeEx FAILED");
			
	INSAppBase* insappbase = static_cast<INSAppBase*>(((void**)lpParameter)[0]);
	MyINSCallbackEnum *myinscallbackenum = static_cast<MyINSCallbackEnum*>(((void**)lpParameter)[1]);
	int a = (int)((void**)lpParameter)[2];
	hr = insappbase->proc7(myinscallbackenum, a);
	insappbase->Release();
	OutputDebugString("insappbase->proc7 done!");
	myinscallbackenum->done = 1;
	SetEvent(myinscallbackenum->record_dispatch_event);
	SetEvent(myinscallbackenum->fieldval_dispatch_event);
	SetEvent(myinscallbackenum->proc5_dispatch_event);
	SetEvent(myinscallbackenum->proc6_dispatch_event);
	SetEvent(myinscallbackenum->fielddef_dispatch_event);
	myinscallbackenum->Release();
	free(lpParameter);
	return hr;
}
DWORD WINAPI callback_instable_EnumFields(LPVOID lpParameter) {
	OutputDebugString("callback_instable_EnumFields");
	HRESULT hr = CoInitializeEx(0, COINIT_MULTITHREADED);
	if (SUCCEEDED(hr)) OutputDebugString("CoInitializeEx SUCCEEDED");
	else OutputDebugString("CoInitializeEx FAILED");
			
	INSTable* instable = static_cast<INSTable*>(((void**)lpParameter)[0]);
	MyINSCallbackEnum *myinscallbackenum = static_cast<MyINSCallbackEnum*>(((void**)lpParameter)[1]);
	int language = (int)((void**)lpParameter)[2];
	hr = instable->EnumFields(myinscallbackenum, language);
	instable->Release();
	OutputDebugString("instable->EnumFields done!");
	myinscallbackenum->done = 1;
	SetEvent(myinscallbackenum->record_dispatch_event);
	SetEvent(myinscallbackenum->fieldval_dispatch_event);
	SetEvent(myinscallbackenum->proc5_dispatch_event);
	SetEvent(myinscallbackenum->proc6_dispatch_event);
	SetEvent(myinscallbackenum->fielddef_dispatch_event);
	myinscallbackenum->Release();
	free(lpParameter);
	OutputDebugString("callback_instable_EnumFields done");
	return hr;
}
DWORD WINAPI callback_instable_proc8(LPVOID lpParameter) {
	OutputDebugString("callback_instable_proc8");
	HRESULT hr = CoInitializeEx(0, COINIT_MULTITHREADED);
	if (SUCCEEDED(hr)) OutputDebugString("CoInitializeEx SUCCEEDED");
	else OutputDebugString("CoInitializeEx FAILED");
			
	INSTable* instable = static_cast<INSTable*>(((void**)lpParameter)[0]);
	MyINSCallbackEnum *myinscallbackenum = static_cast<MyINSCallbackEnum*>(((void**)lpParameter)[1]);
	hr = instable->proc8(myinscallbackenum);
	OutputDebugString("instable->proc8 done!");
	instable->Release();
	myinscallbackenum->done = 1;
	SetEvent(myinscallbackenum->record_dispatch_event);
	SetEvent(myinscallbackenum->fieldval_dispatch_event);
	SetEvent(myinscallbackenum->proc5_dispatch_event);
	SetEvent(myinscallbackenum->proc6_dispatch_event);
	SetEvent(myinscallbackenum->fielddef_dispatch_event);
	myinscallbackenum->Release();
	free(lpParameter);
	return hr;
}
HRESULT _stdcall MyDIRec::EnumFieldValues(IDispatch **en) {
	OutputDebugString("EnumFieldValue");
	MyDIEnum *mydienum = new MyDIEnum;
	mydienum->AddRef();
	*en = mydienum;
	void ** parameters;
	parameters = (void**)malloc(8);
	insrec->AddRef();
	parameters[0] = (void*)insrec;
	mydienum->myinscallbackenum->AddRef();
	parameters[1] = (void*)(mydienum->myinscallbackenum);

	DWORD ThreadId;
	CreateThread(NULL, 1024, callback_insrec_EnumFieldValues, parameters, 0, &ThreadId);
	return S_OK;
}
class MyDIHyperlink : public DIHyperlink {
private:
	long m_cRef;
	ITypeInfo *m_pTypeInfo;
	INSHyperlinkPtr inshyperlinkptr;
public:
	MyDIHyperlink(INSHyperlinkPtr inshyperlinkptr) : m_cRef(0) {
		OutputDebugString("MyDIHyperlink");
		ITypeLib *ptl = 0;
		HRESULT hr = LoadRegTypeLib(LIBID_NavisionAutomation, 1,0,0, &ptl);
                if (SUCCEEDED(hr)) {
			hr = ptl->GetTypeInfoOfGuid(IID_DIHyperlink, &m_pTypeInfo);
			ptl->Release();
		}
		this->inshyperlinkptr = inshyperlinkptr;
	}
	~MyDIHyperlink() {
		OutputDebugString("~MyDIHyperlink");
	}
	HRESULT _stdcall QueryInterface(const IID & riid,void **ppvObject) {
		if (IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_IDispatch) || IsEqualIID(riid, IID_DIHyperlink)) {
			*ppvObject = this;
			reinterpret_cast<IUnknown*>(*ppvObject)->AddRef();
			return NOERROR;
		} else {
			*ppvObject = 0;
			return E_NOINTERFACE;
		}
	}
	ULONG _stdcall AddRef() {
		return InterlockedIncrement(&m_cRef);
	}
	ULONG _stdcall Release() {
		OutputDebugString("MyDIHyperlink::Release");
		LONG res = InterlockedDecrement(&m_cRef);
		if (res == 0) delete this;
		return res;
	}
	HRESULT _stdcall GetTypeInfoCount(UINT * pctinfo) {
		*pctinfo = 1;
		return S_OK;
	}
	HRESULT _stdcall GetTypeInfo(UINT iTInfo, LCID lcid, ITypeInfo ** ppTInfo) {
		(*ppTInfo = m_pTypeInfo)->AddRef();
		return S_OK;
	}
	HRESULT _stdcall GetIDsOfNames(const IID & riid,LPOLESTR *rgszNames,UINT cNames,LCID lcid, DISPID* rgDispId) {
		return m_pTypeInfo->GetIDsOfNames(rgszNames, cNames, rgDispId);
	}
	HRESULT _stdcall Invoke(DISPID dispIdMember,const IID &riid,LCID lcid,WORD wFlags, DISPPARAMS * pDispParams,VARIANT *pVarResult, EXCEPINFO *pExcepInfo, UINT * puArgErr) {
		void *pvThis = static_cast<DIHyperlink*>(this);
		return m_pTypeInfo->Invoke(pvThis, dispIdMember, wFlags, pDispParams, pVarResult, pExcepInfo, puArgErr);
	}
	HRESULT _stdcall Open(BSTR hyperlink) {
		inshyperlinkptr->Open(hyperlink);
		return S_OK;
	}
	HRESULT _stdcall get_proc4(int *a) {
		inshyperlinkptr->proc4(a);
		return S_OK;
	}
	
};
class MyDITable : public DITable {
private:
	long m_cRef;
	ITypeInfo *m_pTypeInfo;
	INSTablePtr instableptr;
public:
	MyDITable(INSTable *instable) : m_cRef(0) {
		OutputDebugString("MyDITable()");
		ITypeLib *ptl = 0;
		HRESULT hr = LoadRegTypeLib(LIBID_NavisionAutomation, 1,0,0, &ptl);
                if (SUCCEEDED(hr)) {
			hr = ptl->GetTypeInfoOfGuid(IID_DITable, &m_pTypeInfo);
			ptl->Release();
		}
		this->instableptr = instable;
	}
	~MyDITable() {
		OutputDebugString("~MyDITable");
	}
	HRESULT _stdcall QueryInterface(const IID & riid,void **ppvObject) {
		if (IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_IDispatch) || IsEqualIID(riid, IID_DITable)) {
			*ppvObject = this;
			reinterpret_cast<IUnknown*>(*ppvObject)->AddRef();
			return NOERROR;
		} else {
			*ppvObject = 0;
			return E_NOINTERFACE;
		}
	}
	ULONG _stdcall AddRef() {
		return InterlockedIncrement(&m_cRef);
	}
	ULONG _stdcall Release() {
		LONG res = InterlockedDecrement(&m_cRef);
		std::stringstream ss;
		ss << "MyDITable::Release (" << res << ")";
		OutputDebugString(ss.str().c_str());
		if (res == 0) delete this;
		return res;
	}
	HRESULT _stdcall GetTypeInfoCount(UINT * pctinfo) {
		*pctinfo = 1;
		return S_OK;
	}
	HRESULT _stdcall GetTypeInfo(UINT iTInfo, LCID lcid, ITypeInfo ** ppTInfo) {
		(*ppTInfo = m_pTypeInfo)->AddRef();
		return S_OK;
	}
	HRESULT _stdcall GetIDsOfNames(const IID & riid,LPOLESTR *rgszNames,UINT cNames,LCID lcid, DISPID* rgDispId) {
		return m_pTypeInfo->GetIDsOfNames(rgszNames, cNames, rgDispId);
	}
	HRESULT _stdcall Invoke(DISPID dispIdMember,const IID &riid,LCID lcid,WORD wFlags, DISPPARAMS * pDispParams,VARIANT *pVarResult, EXCEPINFO *pExcepInfo, UINT * puArgErr) {
		void *pvThis = static_cast<DITable*>(this);
		return m_pTypeInfo->Invoke(pvThis, dispIdMember, wFlags, pDispParams, pVarResult, pExcepInfo, puArgErr);
	}
	HRESULT _stdcall put_PutRec3(IDispatch **a) {
		//TODO: Implement
		return S_OK;
	}
	HRESULT _stdcall put_PutRec4(IDispatch **a) {
		//TODO: Implement
		return S_OK;
	}
	HRESULT _stdcall put_PutRec5(IDispatch **a) {
		//TODO: Implement
		return S_OK;
	}
	HRESULT _stdcall get_GetRec6(IDispatch **a) {
		INSRec *insrec;
		instableptr->proc6(&insrec);
		MyDIRec *mydirec = new MyDIRec(insrec);
		mydirec->AddRef();
		*a = mydirec;
		return S_OK;
	}
	HRESULT _stdcall put_proc7(int a,BSTR *b) {
		instableptr->proc7(a, b);
		return S_OK;
	}
	HRESULT _stdcall get_getDIEnum8(IDispatch **en) {
		MyDIEnum *mydienum = new MyDIEnum;
		mydienum->AddRef();
		*en = mydienum;
		void ** parameters;
		parameters = (void**)malloc(8);
		instableptr->AddRef();
		parameters[0] = (void*)instableptr;
		mydienum->myinscallbackenum->AddRef();
		parameters[1] = (void*)(mydienum->myinscallbackenum);

		DWORD ThreadId;
		CreateThread(NULL, 1024, callback_instable_proc8, parameters, 0, &ThreadId);
		return S_OK;
	}
	HRESULT _stdcall get_EnumRecords(IDispatch **en) {
		OutputDebugString("EnumRecords");
		MyDIEnum *mydienum = new MyDIEnum;
		mydienum->AddRef();
		*en = mydienum;
		void ** parameters;
		parameters = (void**)malloc(8);
		instableptr->AddRef();
		parameters[0] = (void*)instableptr;
		mydienum->myinscallbackenum->AddRef();
		parameters[1] = (void*)(mydienum->myinscallbackenum);

		DWORD ThreadId;
		CreateThread(NULL, 1024, callback_instable_EnumRecords, parameters, 0, &ThreadId);
		return S_OK;

	}
	HRESULT _stdcall get_EnumFields(int language,IDispatch **en) {
		OutputDebugString("EnumFields");
		MyDIEnum *mydienum = new MyDIEnum;
		mydienum->AddRef();
		*en = mydienum;
		void ** parameters;
		parameters = (void**)malloc(12);
		instableptr->AddRef();
		parameters[0] = (void*)instableptr;
		mydienum->myinscallbackenum->AddRef();
		parameters[1] = (void*)(mydienum->myinscallbackenum);
		parameters[2] = (void*)language;

		DWORD ThreadId;
		CreateThread(NULL, 1024, callback_instable_EnumFields, parameters, 0, &ThreadId);
		return S_OK;
	}
	HRESULT _stdcall put_PutRec11(IDispatch **record) {
		//TODO
		return S_OK;
	}
	HRESULT _stdcall get_GetID(int *a) {
		instableptr->GetID(a);
		return S_OK;
	}
	HRESULT _stdcall get_proc13(int *a) {
		instableptr->proc13(a);
		return S_OK;
	}
};
class MyDIMenuButton : public DIMenuButton {
private:
	long m_cRef;
	ITypeInfo *m_pTypeInfo;
	INSMenuButtonPtr insmenubuttonptr;
	IMallocPtr imalloc;
public:
	MyDIMenuButton(INSMenuButtonPtr insmenubuttonptr) : m_cRef(0) {
		OutputDebugString("MyDIMenuButton");
		ITypeLib *ptl = 0;
		HRESULT hr = LoadRegTypeLib(LIBID_NavisionAutomation, 1,0,0, &ptl);
                if (SUCCEEDED(hr)) {
			hr = ptl->GetTypeInfoOfGuid(IID_DIMenuButton, &m_pTypeInfo);
			ptl->Release();
		}
		this->insmenubuttonptr = insmenubuttonptr;
		CoGetMalloc(1, &imalloc);
	}
	~MyDIMenuButton() {
		OutputDebugString("~MyDIButton");
	}
	HRESULT _stdcall QueryInterface(const IID & riid,void **ppvObject) {
		if (IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_IDispatch) || IsEqualIID(riid, IID_DIMenuButton)) {
			*ppvObject = this;
			reinterpret_cast<IUnknown*>(*ppvObject)->AddRef();
			return NOERROR;
		} else {
			*ppvObject = 0;
			return E_NOINTERFACE;
		}
	}
	ULONG _stdcall AddRef() {
		return InterlockedIncrement(&m_cRef);
	}
	ULONG _stdcall Release() {
		OutputDebugString("MyDIMenuButton::Release");
		LONG res = InterlockedDecrement(&m_cRef);
		if (res == 0) delete this;
		return res;
	}
	HRESULT _stdcall GetTypeInfoCount(UINT * pctinfo) {
		*pctinfo = 1;
		return S_OK;
	}
	HRESULT _stdcall GetTypeInfo(UINT iTInfo, LCID lcid, ITypeInfo ** ppTInfo) {
		(*ppTInfo = m_pTypeInfo)->AddRef();
		return S_OK;
	}
	HRESULT _stdcall GetIDsOfNames(const IID & riid,LPOLESTR *rgszNames,UINT cNames,LCID lcid, DISPID* rgDispId) {
		return m_pTypeInfo->GetIDsOfNames(rgszNames, cNames, rgDispId);
	}
	HRESULT _stdcall Invoke(DISPID dispIdMember,const IID &riid,LCID lcid,WORD wFlags, DISPPARAMS * pDispParams,VARIANT *pVarResult, EXCEPINFO *pExcepInfo, UINT * puArgErr) {
		void *pvThis = static_cast<DIMenuButton*>(this);
		return m_pTypeInfo->Invoke(pvThis, dispIdMember, wFlags, pDispParams, pVarResult, pExcepInfo, puArgErr);
	}
	HRESULT _stdcall proc3(BSTR a) {
		insmenubuttonptr->proc3(a);
		return S_OK;
	}
	HRESULT _stdcall proc4(int a,BSTR b,BSTR c) {
		insmenubuttonptr->proc4(a, b, c);
		return S_OK;
	}
};
class MyDIForm : public DIForm {
private:
	long m_cRef;
	ITypeInfo *m_pTypeInfo;
	INSFormPtr insformptr;
	IMallocPtr imalloc;
public:
	MyDIForm(INSFormPtr insformptr) : m_cRef(0) {
		OutputDebugString("MyDIForm");
		ITypeLib *ptl = 0;
		HRESULT hr = LoadRegTypeLib(LIBID_NavisionAutomation, 1,0,0, &ptl);
                if (SUCCEEDED(hr)) {
			hr = ptl->GetTypeInfoOfGuid(IID_DIForm, &m_pTypeInfo);
			ptl->Release();
		}
		this->insformptr = insformptr;
		CoGetMalloc(1, &imalloc);
	}
	~MyDIForm() {
		OutputDebugString("~MyDIForm");
	}
	HRESULT _stdcall QueryInterface(const IID & riid,void **ppvObject) {
		if (IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_IDispatch) || IsEqualIID(riid, IID_DIForm)) {
			*ppvObject = this;
			reinterpret_cast<IUnknown*>(*ppvObject)->AddRef();
			return NOERROR;
		} else {
			*ppvObject = 0;
			return E_NOINTERFACE;
		}
	}
	ULONG _stdcall AddRef() {
		return InterlockedIncrement(&m_cRef);
	}
	ULONG _stdcall Release() {
		OutputDebugString("MyDIForm::Release");
		LONG res = InterlockedDecrement(&m_cRef);
		if (res == 0) delete this;
		return res;
	}
	HRESULT _stdcall GetTypeInfoCount(UINT * pctinfo) {
		*pctinfo = 1;
		return S_OK;
	}
	HRESULT _stdcall GetTypeInfo(UINT iTInfo, LCID lcid, ITypeInfo ** ppTInfo) {
		(*ppTInfo = m_pTypeInfo)->AddRef();
		return S_OK;
	}
	HRESULT _stdcall GetIDsOfNames(const IID & riid,LPOLESTR *rgszNames,UINT cNames,LCID lcid, DISPID* rgDispId) {
		return m_pTypeInfo->GetIDsOfNames(rgszNames, cNames, rgDispId);
	}
	HRESULT _stdcall Invoke(DISPID dispIdMember,const IID &riid,LCID lcid,WORD wFlags, DISPPARAMS * pDispParams,VARIANT *pVarResult, EXCEPINFO *pExcepInfo, UINT * puArgErr) {
		void *pvThis = static_cast<DIForm*>(this);
		return m_pTypeInfo->Invoke(pvThis, dispIdMember, wFlags, pDispParams, pVarResult, pExcepInfo, puArgErr);
	}
	HRESULT _stdcall get_Hyperlink(BSTR *link) {
		LPOLESTR ret;
		insformptr->GetHyperlink(&ret);
		*link = SysAllocString(ret);
		imalloc->Free(ret);
		return S_OK;
	}
	HRESULT _stdcall get_ID(BSTR *id) {
		LPOLESTR ret;
		insformptr->GetID(&ret);
		*id = SysAllocString(ret);
		imalloc->Free(ret);
		return S_OK;
	}
	HRESULT _stdcall get_Rec(IDispatch **rec) {
		INSRec *ret;
		insformptr->GetRec(&ret);
		MyDIRec *mydirec = new MyDIRec(ret);
		mydirec->AddRef();
		*rec = mydirec;
		return S_OK;
	}
	HRESULT _stdcall get_Table(IDispatch **table) {
		INSTable *ret;
		insformptr->GetTable(&ret);
		MyDITable *myditable = new MyDITable(ret);
		myditable->AddRef();
		*table = myditable;
		return S_OK;
	}
	HRESULT _stdcall get_LanguageID(int *id) {
		HRESULT hr = insformptr->GetLanguageID(id);
		return hr;
	}
	HRESULT _stdcall get_proc8(IDispatch **dimenubutton) {
		INSMenuButton *insmenubutton;
		insformptr->proc8(&insmenubutton);
		MyDIMenuButton *mydimenubutton = new MyDIMenuButton(insmenubutton);
		mydimenubutton->AddRef();
		*dimenubutton = mydimenubutton;
		return S_OK;
	}
	HRESULT _stdcall proc9(void) {
		insformptr->proc9();
		return S_OK;
	}
};
class MyDIApplication : public DIApplication {
private:
	long m_cRef;
	ITypeInfo *m_pTypeInfo;
	INSApplicationPtr insapplicationptr;
public:
	MyDIApplication(INSApplicationPtr insapplicationptr) : m_cRef(0) {
		OutputDebugString("MyDIApplication");
		ITypeLib *ptl = 0;
		HRESULT hr = LoadRegTypeLib(LIBID_NavisionAutomation, 1,0,0, &ptl);
                if (SUCCEEDED(hr)) {
			hr = ptl->GetTypeInfoOfGuid(IID_DIApplication, &m_pTypeInfo);
			ptl->Release();
		}
		this->insapplicationptr = insapplicationptr;
	}
	~MyDIApplication() {
		OutputDebugString("~MyDIApplication");
	}
	HRESULT _stdcall QueryInterface(const IID & riid,void **ppvObject) {
		if (IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_IDispatch) || IsEqualIID(riid, IID_DIApplication)) {
			*ppvObject = this;
			reinterpret_cast<IUnknown*>(*ppvObject)->AddRef();
			return NOERROR;
		} else {
			*ppvObject = 0;
			return E_NOINTERFACE;
		}
	}
	ULONG _stdcall AddRef() {
		return InterlockedIncrement(&m_cRef);
	}
	ULONG _stdcall Release() {
		OutputDebugString("MyDIApplication::Release");
		LONG res = InterlockedDecrement(&m_cRef);
		if (res == 0) delete this;
		return res;
	}
	HRESULT _stdcall GetTypeInfoCount(UINT * pctinfo) {
		*pctinfo = 1;
		return S_OK;
	}
	HRESULT _stdcall GetTypeInfo(UINT iTInfo, LCID lcid, ITypeInfo ** ppTInfo) {
		(*ppTInfo = m_pTypeInfo)->AddRef();
		return S_OK;
	}
	HRESULT _stdcall GetIDsOfNames(const IID & riid,LPOLESTR *rgszNames,UINT cNames,LCID lcid, DISPID* rgDispId) {
		return m_pTypeInfo->GetIDsOfNames(rgszNames, cNames, rgDispId);
	}
	HRESULT _stdcall Invoke(DISPID dispIdMember,const IID &riid,LCID lcid,WORD wFlags, DISPPARAMS * pDispParams,VARIANT *pVarResult, EXCEPINFO *pExcepInfo, UINT * puArgErr) {
		void *pvThis = static_cast<DIApplication*>(this);
		return m_pTypeInfo->Invoke(pvThis, dispIdMember, wFlags, pDispParams, pVarResult, pExcepInfo, puArgErr);
	}
	HRESULT _stdcall get_Form(IDispatch **form) {
		INSForm *insform;
		HRESULT hr = insapplicationptr->GetForm(&insform);
		MyDIForm *mydiform = new MyDIForm(insform);
		mydiform->AddRef();
		*form = mydiform;
		return hr;
	}
};
class MyDIAppBase : public DIAppBase {
private:
	long m_cRef;
	ITypeInfo *m_pTypeInfo;
	INSAppBasePtr insappbase;
public:
	MyDIAppBase(INSAppBasePtr insappbase) : m_cRef(0) {
		ITypeLib *ptl = 0;
		HRESULT hr = LoadRegTypeLib(LIBID_NavisionAutomation, 1,0,0, &ptl);
                if (SUCCEEDED(hr)) {
			hr = ptl->GetTypeInfoOfGuid(IID_DIAppBase, &m_pTypeInfo);
			ptl->Release();
		}
		this->insappbase = insappbase;
	}
	~MyDIAppBase() {
		OutputDebugString("~MyDIAppBase");
	}
	HRESULT _stdcall QueryInterface(const IID & riid,void **ppvObject) {
		if (IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_IDispatch) || IsEqualIID(riid, IID_DIAppBase)) {
			*ppvObject = this;
			reinterpret_cast<IUnknown*>(*ppvObject)->AddRef();
			return NOERROR;
		} else {
			*ppvObject = 0;
			return E_NOINTERFACE;
		}
	}
	ULONG _stdcall AddRef() {
		return InterlockedIncrement(&m_cRef);
	}
	ULONG _stdcall Release() {
		LONG res = InterlockedDecrement(&m_cRef);
		std::stringstream ss;
		ss << "MyDIAppBase::Release (" << res << ")";
		OutputDebugString(ss.str().c_str());
		if (res == 0) delete this;
		return res;
	}
	HRESULT _stdcall GetTypeInfoCount(UINT * pctinfo) {
		*pctinfo = 1;
		return S_OK;
	}
	HRESULT _stdcall GetTypeInfo(UINT iTInfo, LCID lcid, ITypeInfo ** ppTInfo) {
		(*ppTInfo = m_pTypeInfo)->AddRef();
		return S_OK;
	}
	HRESULT _stdcall GetIDsOfNames(const IID & riid,LPOLESTR *rgszNames,UINT cNames,LCID lcid, DISPID* rgDispId) {
		return m_pTypeInfo->GetIDsOfNames(rgszNames, cNames, rgDispId);
	}
	HRESULT _stdcall Invoke(DISPID dispIdMember,const IID &riid,LCID lcid,WORD wFlags, DISPPARAMS * pDispParams,VARIANT *pVarResult, EXCEPINFO *pExcepInfo, UINT * puArgErr) {
		void *pvThis = static_cast<DIAppBase*>(this);
		return m_pTypeInfo->Invoke(pvThis, dispIdMember, wFlags, pDispParams, pVarResult, pExcepInfo, puArgErr);
	}
	HRESULT _stdcall GetTable(int tableId,IDispatch **table) {
		OutputDebugString("GetTable");
		INSTable *instable;
		insappbase->GetTable(tableId, &instable);
		*table = new MyDITable(instable);
		(*table)->AddRef();
		return S_OK;
	}
	HRESULT _stdcall get_GetInfo(SAFEARRAY **info) {
		*info = SafeArrayCreateVector(VT_BSTR, 0, 4);
		BSTR *infoarray;
		SafeArrayAccessData(*info, (void**)&infoarray);
		insappbase->GetInfo(&infoarray[0], &infoarray[1], &infoarray[2], &infoarray[3]);
		SafeArrayUnaccessData(*info);
		return S_OK;
	}
	HRESULT _stdcall proc5() {
		insappbase->proc5();
		return S_OK;
	}
	HRESULT _stdcall proc6(int a) {
		insappbase->proc6(a);
		return S_OK;
	}
	HRESULT _stdcall get_getEnum7(int b,IDispatch **dienum) {
		OutputDebugString("getEnum7");
		MyDIEnum *mydienum = new MyDIEnum;
		mydienum->AddRef();
		*dienum = mydienum;
		void ** parameters;
		parameters = (void**)malloc(12);
		insappbase->AddRef();
		parameters[0] = (void*)insappbase;
		mydienum->myinscallbackenum->AddRef();
		parameters[1] = (void*)(mydienum->myinscallbackenum);
		parameters[2] = (void*)b;

		DWORD ThreadId;
		OutputDebugString("CreateThread");
		CreateThread(NULL, 1024, callback_insappbase_proc7, parameters, 0, &ThreadId);
		OutputDebugString("getEnum7 done");
		return S_OK;
	}
};
class MyDIObjectDesigner : public DIObjectDesigner {
private:
	long m_cRef;
	ITypeInfo *m_pTypeInfo;
	INSObjectDesignerPtr insobjectdesigner;
	IMallocPtr imalloc;
public:
	MyDIObjectDesigner(INSObjectDesignerPtr insobjectdesigner) : m_cRef(0) {
		OutputDebugString("MyDIObjectDesigner");
		ITypeLib *ptl = 0;
		HRESULT hr = LoadRegTypeLib(LIBID_NavisionAutomation, 1,0,0, &ptl);
                if (SUCCEEDED(hr)) {
			hr = ptl->GetTypeInfoOfGuid(IID_DIObjectDesigner, &m_pTypeInfo);
			ptl->Release();
		}
		this->insobjectdesigner = insobjectdesigner;
		CoGetMalloc(1, &imalloc);
	}
	~MyDIObjectDesigner() {
		OutputDebugString("~MyDIObjectDesigner");
	}
	HRESULT _stdcall QueryInterface(const IID & riid,void **ppvObject) {
		if (IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_IDispatch) || IsEqualIID(riid, IID_DIObjectDesigner)) {
			*ppvObject = this;
			reinterpret_cast<IUnknown*>(*ppvObject)->AddRef();
			return NOERROR;
		} else {
			*ppvObject = 0;
			return E_NOINTERFACE;
		}
	}
	ULONG _stdcall AddRef() {
		return InterlockedIncrement(&m_cRef);
	}
	ULONG _stdcall Release() {
		OutputDebugString("MyDIObjectDesigner::Release");
		LONG res = InterlockedDecrement(&m_cRef);
		if (res == 0) delete this;
		return res;
	}
	HRESULT _stdcall GetTypeInfoCount(UINT * pctinfo) {
		*pctinfo = 1;
		return S_OK;
	}
	HRESULT _stdcall GetTypeInfo(UINT iTInfo, LCID lcid, ITypeInfo ** ppTInfo) {
		(*ppTInfo = m_pTypeInfo)->AddRef();
		return S_OK;
	}
	HRESULT _stdcall GetIDsOfNames(const IID & riid,LPOLESTR *rgszNames,UINT cNames,LCID lcid, DISPID* rgDispId) {
		return m_pTypeInfo->GetIDsOfNames(rgszNames, cNames, rgDispId);
	}
	HRESULT _stdcall Invoke(DISPID dispIdMember,const IID &riid,LCID lcid,WORD wFlags, DISPPARAMS * pDispParams,VARIANT *pVarResult, EXCEPINFO *pExcepInfo, UINT * puArgErr) {
		void *pvThis = static_cast<DIObjectDesigner*>(this);
		return m_pTypeInfo->Invoke(pvThis, dispIdMember, wFlags, pDispParams, pVarResult, pExcepInfo, puArgErr);
	}
	HRESULT _stdcall get_ReadObject(int objectType,int objectId, IDispatch ** destination) {
		IStreamPtr strm;
		HRESULT hr;
		CreateStreamOnHGlobal(NULL, TRUE, &strm);
		hr = insobjectdesigner->ReadObject(objectType, objectId, strm);
		if (SUCCEEDED(hr)) {
			*destination = new MyDIStream(strm);
			(*destination)->AddRef();
		}
		return S_OK;
	}
	HRESULT _stdcall get_ReadObjects(BSTR filter,IDispatch **destination) {
		OutputDebugString("ReadObjects");
		IStreamPtr strm;
		CreateStreamOnHGlobal(NULL, TRUE, &strm);
		if (SUCCEEDED(insobjectdesigner->ReadObjects(filter, strm))) {
			*destination = new MyDIStream(strm);
			(*destination)->AddRef();
		}
		return S_OK;
	}
	HRESULT _stdcall get_WriteObjects(IDispatch **source) {
		OutputDebugString("WriteObjects");
		IStream* strm;
		CreateStreamOnHGlobal(NULL, TRUE, &strm);
		if (SUCCEEDED(insobjectdesigner->WriteObjects(strm))) {
			MyDIStream *mydistream = new MyDIStream(strm);
			mydistream->run_func_free = &mydistream_run_func_free_writeobjects;
			mydistream->run_func = &mydistream_run_func_writeobjects;
			void **run_func_param_array = (void**)malloc(2*4);
			run_func_param_array[0] = insobjectdesigner;
			insobjectdesigner->AddRef();
			run_func_param_array[1] = strm;
			strm->AddRef();
			if (mydistream->run_func_param != 0) free(mydistream->run_func_param);
			mydistream->run_func_param = run_func_param_array;
			mydistream->AddRef();
			*source = mydistream;
		}
		return S_OK;
	}
	HRESULT _stdcall CompileObject(int objectType,int objectId) {
		OutputDebugString("CompileObject");
		insobjectdesigner->CompileObject(objectType, objectId);
		return S_OK;
	}
	HRESULT _stdcall CompileObjects(BSTR filter) {
		OutputDebugString("CompileObjects");
		insobjectdesigner->CompileObjects(filter);
		return S_OK;
	}
	HRESULT _stdcall get_GetServerName(BSTR *serverName) {
		LPOLESTR ret;
		OutputDebugString("GetServerName");
		insobjectdesigner->GetServerName(&ret);
		*serverName = SysAllocString(ret);
		imalloc->Free(ret);
		return S_OK;
	}
	HRESULT _stdcall get_GetDatabaseName(BSTR *databaseName) {
		LPOLESTR ret;
		OutputDebugString("GetDatabaseName");
		insobjectdesigner->GetDatabaseName(&ret);
		*databaseName = SysAllocString(ret);
		imalloc->Free(ret);
		return S_OK;
	}
	HRESULT _stdcall get_GetServerType(int *serverType) {
		OutputDebugString("GetServerType");
		insobjectdesigner->GetServerType(serverType);
		return S_OK;
	}
	HRESULT _stdcall get_GetCSIDEVersion(BSTR *csideVersion) {
		LPOLESTR ret;
		OutputDebugString("GetCSIDEVersion");
		insobjectdesigner->GetCSIDEVersion(&ret);
		*csideVersion = SysAllocString(ret);
		imalloc->Free(ret);
		return S_OK;
	}
	HRESULT _stdcall get_GetApplicationVersion(BSTR *applicationVersion) {
		LPOLESTR ret;
		OutputDebugString("GetApplicationVersion");
		insobjectdesigner->GetApplicationVersion(&ret);
		*applicationVersion = SysAllocString(ret);
		imalloc->Free(ret);
		return S_OK;
	}
	HRESULT _stdcall get_GetCompanyName(BSTR *companyName) {
		HRESULT hr;
		LPOLESTR ret;
		OutputDebugString("GetCompanyName");
		hr = insobjectdesigner->GetCompanyName(&ret);
		if (SUCCEEDED(hr)) OutputDebugString("GetCompanyName OK");
		else OutputDebugString("GetCompanyName NOT OK");
		*companyName = SysAllocString(ret);
		imalloc->Free(ret);
		return S_OK;
	}
};
class MyDIRunningObjectTable : public DIRunningObjectTable {
private:
	long m_cRef;
	ITypeInfo *m_pTypeInfo;
	IRunningObjectTablePtr rot;
	std::vector<IMonikerPtr> monikers;
public:
	MyDIRunningObjectTable(IRunningObjectTablePtr rot) : m_cRef(0) {
		OutputDebugString("MyDIRunningObjectTable");
		ITypeLib *ptl = 0;
		HRESULT hr = LoadRegTypeLib(LIBID_NavisionAutomation, 1,0,0, &ptl);
                if (SUCCEEDED(hr)) {
			hr = ptl->GetTypeInfoOfGuid(IID_DIRunningObjectTable, &m_pTypeInfo);
			ptl->Release();
		}
		this->rot = rot;

		IEnumMonikerPtr enumMoniker;
		hr = rot->EnumRunning(&enumMoniker);
		while (1) {
			IMoniker *moniker;
			hr = enumMoniker->Next(1, &moniker, NULL);
			if (hr == S_FALSE) break;

			if (hr == S_OK) {
				monikers.push_back(moniker);
			}
		}
	}
	~MyDIRunningObjectTable() {
		OutputDebugString("~MyDIRunningObjectTable");
	}
	HRESULT _stdcall QueryInterface(const IID & riid,void **ppvObject) {
		if (IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_IDispatch) || IsEqualIID(riid, IID_DIRunningObjectTable)) {
			*ppvObject = this;
			reinterpret_cast<IUnknown*>(*ppvObject)->AddRef();
			return NOERROR;
		} else {
			*ppvObject = 0;
			return E_NOINTERFACE;
		}
	}
	ULONG _stdcall AddRef() {
		OutputDebugString("RunningObjectTable::AddRef");
		return InterlockedIncrement(&m_cRef);
	}
	ULONG _stdcall Release() {
		LONG res = InterlockedDecrement(&m_cRef);
		std::stringstream ss;
		ss << "RunningObjectTable::Release (" << res << ")";
		OutputDebugString(ss.str().c_str());
		if (res == 0) delete this;
		return res;
	}
	HRESULT _stdcall GetTypeInfoCount(UINT * pctinfo) {
		*pctinfo = 1;
		return S_OK;
	}
	HRESULT _stdcall GetTypeInfo(UINT iTInfo, LCID lcid, ITypeInfo ** ppTInfo) {
		(*ppTInfo = m_pTypeInfo)->AddRef();
		return S_OK;
	}
	HRESULT _stdcall GetIDsOfNames(const IID & riid,LPOLESTR *rgszNames,UINT cNames,LCID lcid, DISPID* rgDispId) {
		return m_pTypeInfo->GetIDsOfNames(rgszNames, cNames, rgDispId);
	}
	HRESULT _stdcall Invoke(DISPID dispIdMember,const IID &riid,LCID lcid,WORD wFlags, DISPPARAMS * pDispParams,VARIANT *pVarResult, EXCEPINFO *pExcepInfo, UINT * puArgErr) {
		void *pvThis = static_cast<DIRunningObjectTable*>(this);
		return m_pTypeInfo->Invoke(pvThis, dispIdMember, wFlags, pDispParams, pVarResult, pExcepInfo, puArgErr);
	}
	HRESULT _stdcall get_Names(SAFEARRAY **names) {
		OutputDebugString("get_Names");
		*names = SafeArrayCreateVector(VT_BSTR, 0, monikers.size());
		BSTR *namesarray;
		SafeArrayAccessData(*names, (void**)&namesarray);
		IBindCtxPtr bc;
		CreateBindCtx(0, &bc);
		IMallocPtr imalloc;
		CoGetMalloc(1, &imalloc);
		for (unsigned int i = 0; i < monikers.size(); i++) {
			LPOLESTR displayname;
			monikers[i]->GetDisplayName(bc, 0, &displayname);
			namesarray[i] = SysAllocString(displayname);
			imalloc->Free(displayname);
		}
		SafeArrayUnaccessData(*names);
		return S_OK;
	}
	HRESULT _stdcall get_ObjectDesigner(unsigned int index,IDispatch **diobjectdesigner) {
		OutputDebugString("getObjectDesigner");
		HRESULT hr;
		INSObjectDesignerPtr insobjectdesigner;
		IUnknownPtr iunknown;
		if ((index < 0) || (index >= monikers.size())) return S_OK;
		hr = rot->GetObject(monikers[index], &iunknown);
		hr = iunknown->QueryInterface(IID_INSObjectDesigner, (void**)&insobjectdesigner);
		*diobjectdesigner = new MyDIObjectDesigner(insobjectdesigner);
		(*diobjectdesigner)->AddRef();
		return S_OK;
	}
	HRESULT _stdcall get_AppBase(unsigned int index,IDispatch **diappbase) {
		OutputDebugString("getAppBase");
		HRESULT hr;
		INSAppBasePtr insappbase;
		IUnknownPtr iunknown;
		if ((index < 0) || (index >= monikers.size())) return S_OK;
		hr = rot->GetObject(monikers[index], &iunknown);
		hr = iunknown->QueryInterface(IID_INSAppBase, (void**)&insappbase);
		*diappbase = new MyDIAppBase(insappbase);
		(*diappbase)->AddRef();
		return S_OK;
	}
	HRESULT _stdcall get_Hyperlink(unsigned int index,IDispatch **hyperlink) {
		OutputDebugString("getHyperlink");
		HRESULT hr;
		INSHyperlinkPtr inshyperlinkptr;
		IUnknownPtr iunknown;
		if ((index < 0) || (index >= monikers.size())) return S_OK;
		hr = rot->GetObject(monikers[index], &iunknown);
		hr = iunknown->QueryInterface(IID_INSHyperlink, (void**)&inshyperlinkptr);
		*hyperlink = new MyDIHyperlink(inshyperlinkptr);
		(*hyperlink)->AddRef();
		return S_OK;
	}
	HRESULT _stdcall get_Application(unsigned int index,IDispatch **application) {
		OutputDebugString("getApplication");
		HRESULT hr;
		INSApplicationPtr insapplicationptr;
		IUnknownPtr iunknown;
		if ((index < 0) || (index >= monikers.size())) return S_OK;
		hr = rot->GetObject(monikers[index], &iunknown);
		hr = iunknown->QueryInterface(IID_INSApplication, (void**)&insapplicationptr);
		if (!SUCCEEDED(hr)) OutputDebugString("INSApplication failed");
		*application = new MyDIApplication(insapplicationptr);
		(*application)->AddRef();
		return S_OK;
	}
};
class MyNavisionAutomation : public NavisionAutomation {
private:
	long m_cRef;
	ITypeInfo *m_pTypeInfo;
public:
	MyNavisionAutomation() : m_cRef(0) {
		OutputDebugString("MyNavisionAutomation");
		ITypeLib *ptl = 0;
		HRESULT hr = LoadRegTypeLib(LIBID_NavisionAutomation, 1,0,0, &ptl);
                if (SUCCEEDED(hr)) {
			hr = ptl->GetTypeInfoOfGuid(IID_NavisionAutomation, &m_pTypeInfo);
			ptl->Release();
		}
	}
	~MyNavisionAutomation() {
		OutputDebugString("~MyNavisionAutomation");
	}
	HRESULT _stdcall QueryInterface(const IID & riid,void **ppvObject) {
		if (IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_IDispatch) || IsEqualIID(riid, IID_NavisionAutomation)) {
			*ppvObject = this;
			reinterpret_cast<IUnknown*>(*ppvObject)->AddRef();
			return NOERROR;
		} else {
			*ppvObject = 0;
			return E_NOINTERFACE;
		}
	}
	ULONG _stdcall AddRef() {
		OutputDebugString("MyNavisionAutomation::AddRef");
		return InterlockedIncrement(&m_cRef);
	}
	ULONG _stdcall Release() {
		LONG res = InterlockedDecrement(&m_cRef);
		std::stringstream ss;
		ss << "MyNavisionAutomation::Release (" << res << ")";
		OutputDebugString(ss.str().c_str());
		if (res == 0) delete this;
		return res;
	}
	HRESULT _stdcall GetTypeInfoCount(UINT * pctinfo) {
		*pctinfo = 1;
		return S_OK;
	}
	HRESULT _stdcall GetTypeInfo(UINT iTInfo, LCID lcid, ITypeInfo ** ppTInfo) {
		(*ppTInfo = m_pTypeInfo)->AddRef();
		return S_OK;
	}
	HRESULT _stdcall GetIDsOfNames(const IID & riid,LPOLESTR *rgszNames,UINT cNames,LCID lcid, DISPID* rgDispId) {
		return m_pTypeInfo->GetIDsOfNames(rgszNames, cNames, rgDispId);
	}
	HRESULT _stdcall Invoke(DISPID dispIdMember,const IID &riid,LCID lcid,WORD wFlags, DISPPARAMS * pDispParams,VARIANT *pVarResult, EXCEPINFO *pExcepInfo, UINT * puArgErr) {
		void *pvThis = static_cast<NavisionAutomation*>(this);
		return m_pTypeInfo->Invoke(pvThis, dispIdMember, wFlags, pDispParams, pVarResult, pExcepInfo, puArgErr);
	}
	HRESULT _stdcall get_RunningObjectTable(IDispatch **dirunningobjecttable) {
		IRunningObjectTablePtr rot;
		GetRunningObjectTable(0, &rot);
		*dirunningobjecttable = new MyDIRunningObjectTable(rot);
		(*dirunningobjecttable)->AddRef();
		return S_OK;
	}
};

class MyIClassFactory : public IClassFactory {
public:
	//IUnknown
	HRESULT __stdcall QueryInterface(const IID& iid, void** ppvObject) {
		if ((iid == IID_IUnknown) || (iid == IID_IClassFactory)) {
			*ppvObject = static_cast<IClassFactory*>(this);
			reinterpret_cast<IUnknown*>(*ppvObject)->AddRef();
		} else {
			*ppvObject = NULL;
			return E_NOINTERFACE;
		}
		AddRef();
		return S_OK;
	}
	ULONG __stdcall AddRef() {
		return InterlockedIncrement(&m_cRef);
	}
	ULONG __stdcall Release() {
		LONG res = InterlockedDecrement(&m_cRef);
		std::stringstream ss;
		ss << "MyIClassFactory::Release (" << res << ")";
		OutputDebugString(ss.str().c_str());
		if (res == 0) delete this;
		return res;
	}
	//IClassFactory
	HRESULT __stdcall CreateInstance(IUnknown* pUnknownOuter, const IID& iid, void **ppv) {
		HRESULT hr;
		if (pUnknownOuter != NULL) return CLASS_E_NOAGGREGATION;
		if ((iid == IID_NavisionAutomation) || (iid == IID_IUnknown) || (iid == IID_IDispatch)) {
			MyNavisionAutomation *myNavisionAutomation = new MyNavisionAutomation;
			myNavisionAutomation->AddRef();
			hr = myNavisionAutomation->QueryInterface(iid, ppv);
			myNavisionAutomation->Release();
		} else {
			*ppv = 0;
			std::wstringstream wss;
			wss << "CLASS_E_CLASSNOTAVAILABLE ";
			LPOLESTR lpsz;
			StringFromCLSID(iid, &lpsz);
			wss << lpsz;
	                OutputDebugStringW(wss.str().c_str());
			hr = CLASS_E_CLASSNOTAVAILABLE;
		}
		return hr;
	}
	HRESULT __stdcall LockServer(BOOL bLock) {
		static long s_cServerLocks = 0;
		if (bLock) InterlockedIncrement(&s_cServerLocks);
		else InterlockedDecrement(&s_cServerLocks);
		return S_OK;
	}
	MyIClassFactory() : m_cRef(0) {
	}
private:
	long m_cRef;
};
STDAPI DllGetClassObject(REFCLSID rclsid, REFIID riid, LPVOID *ppv) {
	OutputDebugString("DllGetClassObject");
	HRESULT hr;
	MyIClassFactory *myIClassFactory = new MyIClassFactory;
	hr = myIClassFactory->QueryInterface(riid, ppv);
	return hr;
}
STDAPI DllCanUnloadNow() {
	OutputDebugString("DllCanUnloadNow");
	return S_FALSE;
}
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved) {
	g_hinstDLL = hinstDLL;
	OutputDebugString("DllMain");
	switch (fdwReason) {
	case DLL_PROCESS_ATTACH:
		OutputDebugString("Process Attach");
		break;
	case DLL_PROCESS_DETACH:
		OutputDebugString("Process Detach");
		break;
	case DLL_THREAD_ATTACH:
		OutputDebugString("Thread Attach");
		break;
	}
	return TRUE;
}
